# -*- coding: utf-8 -*-
import pytest

from inspect import getmro
from collections import OrderedDict

from mock import Mock
from pytest import raises

from pyee import BaseEventEmitter, EventEmitter


class PyeeTestException(Exception):
    pass


@pytest.mark.parametrize('cls', [
    BaseEventEmitter,
    EventEmitter
])
def test_emit_sync(cls):
    """Basic synchronous emission works"""

    call_me = Mock()
    ee = cls()

    @ee.on('event')
    def event_handler(data, **kwargs):
        call_me()
        assert data == 'emitter is emitted!'

    # Making sure data is passed propers
    ee.emit('event', 'emitter is emitted!', error=False)

    call_me.assert_called_once()


@pytest.mark.parametrize('cls', [
    BaseEventEmitter,
    EventEmitter
])
def test_emit_error(cls):
    """Errors raise with no event handler, otherwise emit on handler"""

    call_me = Mock()
    ee = cls()

    test_exception = PyeeTestException('lololol')

    with raises(PyeeTestException):
        ee.emit('error', test_exception)

    @ee.on('error')
    def on_error(exc):
        call_me()

    # No longer raises and error instead return True indicating handled
    assert ee.emit('error', test_exception) is True
    call_me.assert_called_once()


def test_emit_return():
    """Emit returns True when handlers are registered on an event, and false
    otherwise.
    """

    call_me = Mock()
    ee = BaseEventEmitter()

    # make sure emitting without a callback returns False
    assert not ee.emit('data')

    # add a callback
    ee.on('data')(call_me)

    # should return True now
    assert ee.emit('data')


def test_new_listener_event():
    """The 'new_listener' event fires whenever a new listerner is added."""

    call_me = Mock()
    ee = BaseEventEmitter()

    ee.on('new_listener', call_me)

    # Should fire new_listener event
    @ee.on('event')
    def event_handler(data):
        pass

    call_me.assert_called_once_with('event', event_handler)


def test_listener_removal():
    """Removing listeners removes the correct listener from an event."""

    ee = BaseEventEmitter()

    # Some functions to pass to the EE
    def first():
        return 1

    ee.on('event', first)

    @ee.on('event')
    def second():
        return 2

    @ee.on('event')
    def third():
        return 3

    def fourth():
        return 4

    ee.on('event', fourth)

    assert ee._events['event'] == OrderedDict([
        (first, first),
        (second, second),
        (third, third),
        (fourth, fourth)
    ])

    ee.remove_listener('event', second)

    assert ee._events['event'] == OrderedDict([
        (first, first),
        (third, third),
        (fourth, fourth)
    ])

    ee.remove_listener('event', first)
    assert ee._events['event'] == OrderedDict([
        (third, third),
        (fourth, fourth)
    ])

    ee.remove_all_listeners('event')
    assert ee._events['event'] == OrderedDict()


def test_listener_removal_on_emit():
    """Test that a listener removed during an emit is called inside the current
    emit cycle.
    """

    call_me = Mock()
    ee = BaseEventEmitter()

    def should_remove():
        ee.remove_listener('remove', call_me)

    ee.on('remove', should_remove)
    ee.on('remove', call_me)

    ee.emit('remove')

    call_me.assert_called_once()

    call_me.reset_mock()

    # Also test with the listeners added in the opposite order
    ee = BaseEventEmitter()
    ee.on('remove', call_me)
    ee.on('remove', should_remove)

    ee.emit('remove')

    call_me.assert_called_once()


def test_once():
    """Test that `once()` method works propers.
    """

    # very similar to "test_emit" but also makes sure that the event
    # gets removed afterwards

    call_me = Mock()
    ee = BaseEventEmitter()

    def once_handler(data):
        assert data == 'emitter is emitted!'
        call_me()

    # Tests to make sure that after event is emitted that it's gone.
    ee.once('event', once_handler)

    ee.emit('event', 'emitter is emitted!')

    call_me.assert_called_once()

    assert ee._events['event'] == OrderedDict()


def test_once_removal():
    """Removal of once functions works
    """

    ee = BaseEventEmitter()

    def once_handler(data):
        pass

    handle = ee.once('event', once_handler)

    assert handle == once_handler

    ee.remove_listener('event', handle)

    assert ee._events['event'] == OrderedDict()


def test_listeners():
    """`listeners()` returns a copied list of listeners."""

    call_me = Mock()
    ee = BaseEventEmitter()

    @ee.on('event')
    def event_handler():
        pass

    @ee.once('event')
    def once_handler():
        pass

    listeners = ee.listeners('event')

    assert listeners[0] == event_handler
    assert listeners[1] == once_handler

    # listeners is a copy, you can't mutate the innards this way
    listeners[0] = call_me

    ee.emit('event')

    call_me.assert_not_called()


def test_properties_preserved():
    """Test that the properties of decorated functions are preserved."""

    call_me = Mock()
    call_me_also = Mock()
    ee = BaseEventEmitter()

    @ee.on('always')
    def always_event_handler():
        """An event handler."""
        call_me()

    @ee.once('once')
    def once_event_handler():
        """Another event handler."""
        call_me_also()

    assert always_event_handler.__doc__ == 'An event handler.'
    assert once_event_handler.__doc__ == 'Another event handler.'

    always_event_handler()
    call_me.assert_called_once()

    once_event_handler()
    call_me_also.assert_called_once()

    call_me_also.reset_mock()

    # Calling the event handler directly doesn't clear the handler
    ee.emit('once')
    call_me_also.assert_called_once()


def test_inheritance():
    """Test that inheritance is preserved from object"""
    assert object in getmro(BaseEventEmitter)

    class example(BaseEventEmitter):
        def __init__(self):
            super(example, self).__init__()

    assert BaseEventEmitter in getmro(example)
    assert object in getmro(example)


def test_multiple_inheritance():
    """Test that inheritance is preserved along a lengthy MRO"""
    class example(BaseEventEmitter):
        def __init__(self):
            super(example, self).__init__()

    class _example(example):
        def __init__(self):
            super(_example, self).__init__()

    class example2(_example):
        def __init__(self):
            super(example2, self).__init__()

    class _example2(_example):
        def __init__(self):
            super(_example2, self).__init__()

    a = _example2()  # noqa



# -*- coding: utf-8 -*-

from sys import version_info as v

collect_ignore = []

if not (v[0] >= 3 and v[1] >= 5):
    collect_ignore.append('test_async.py')

if not (v[0] >= 3 and v[1] >= 7):
    collect_ignore.append('test_trio.py')



# -*- coding: utf-8 -*-

import pytest

from mock import call, Mock
from pyee import BaseEventEmitter
from pyee.uplift import uplift


class UpliftedEventEmitter(BaseEventEmitter):
    pass


def test_uplift_emit():
    call_me = Mock()

    base_ee = BaseEventEmitter()

    @base_ee.on('base_event')
    def base_handler():
        call_me('base event on base emitter')

    @base_ee.on('shared_event')
    def shared_base_handler():
        call_me('shared event on base emitter')

    uplifted_ee = uplift(UpliftedEventEmitter, base_ee)

    assert isinstance(
        uplifted_ee,
        UpliftedEventEmitter
    ), 'Returns an uplifted emitter'

    @uplifted_ee.on('uplifted_event')
    def uplifted_handler():
        call_me('uplifted event on uplifted emitter')

    @uplifted_ee.on('shared_event')
    def shared_uplifted_handler():
        call_me('shared event on uplifted emitter')

    # Events on uplifted proxy correctly
    assert uplifted_ee.emit('base_event')
    assert uplifted_ee.emit('shared_event')
    assert uplifted_ee.emit('uplifted_event')

    call_me.assert_has_calls([
        call('base event on base emitter'),
        call('shared event on uplifted emitter'),
        call('shared event on base emitter'),
        call('uplifted event on uplifted emitter')
    ])

    call_me.reset_mock()

    # Events on underlying proxy correctly
    assert base_ee.emit('base_event')
    assert base_ee.emit('shared_event')
    assert base_ee.emit('uplifted_event')

    call_me.assert_has_calls([
        call('base event on base emitter'),
        call('shared event on base emitter'),
        call('shared event on uplifted emitter'),
        call('uplifted event on uplifted emitter')
    ])

    call_me.reset_mock()

    # Quick check for unwrap
    uplifted_ee.unwrap()

    with pytest.raises(AttributeError):
        getattr(uplifted_ee, 'unwrap')

    with pytest.raises(AttributeError):
        getattr(base_ee, 'unwrap')

    assert not uplifted_ee.emit('base_event')
    assert uplifted_ee.emit('shared_event')
    assert uplifted_ee.emit('uplifted_event')

    assert base_ee.emit('base_event')
    assert base_ee.emit('shared_event')
    assert not base_ee.emit('uplifted_event')

    call_me.assert_has_calls([
        # No listener for base event on uplifted
        call('shared event on uplifted emitter'),
        call('uplifted event on uplifted emitter'),
        call('base event on base emitter'),
        call('shared event on base emitter')
        # No listener for uplifted event on uplifted
    ])


@pytest.mark.parametrize('error_handling', [
  'new', 'underlying', 'neither'
])
def test_exception_handling(error_handling):
    base_ee = BaseEventEmitter()
    uplifted_ee = uplift(
        UpliftedEventEmitter, base_ee,
        error_handling=error_handling
    )

    # Exception handling alwyas prefers uplifted
    base_error = Exception('base error')
    uplifted_error = Exception('uplifted error')

    # Hold my beer
    base_error_handler = Mock()
    base_ee._emit_handle_potential_error = base_error_handler

    # Hold my other beer
    uplifted_error_handler = Mock()
    uplifted_ee._emit_handle_potential_error = uplifted_error_handler

    base_ee.emit('error', base_error)
    uplifted_ee.emit('error', uplifted_error)

    if error_handling == 'new':
        base_error_handler.assert_not_called()
        uplifted_error_handler.assert_has_calls([
            call('error', base_error),
            call('error', uplifted_error)
        ])
    elif error_handling == 'underlying':
        base_error_handler.assert_has_calls([
            call('error', base_error),
            call('error', uplifted_error)
        ])
        uplifted_error_handler.assert_not_called()
    elif error_handling == 'neither':
        base_error_handler.assert_called_once_with('error', base_error)
        uplifted_error_handler.assert_called_once_with('error', uplifted_error)
    else:
        raise Exception('unrecognized setting')


@pytest.mark.parametrize('proxy_new_listener', [
    'both',
    'neither',
    'forward',
    'backward'
])
def test_proxy_new_listener(proxy_new_listener):
    call_me = Mock()

    base_ee = BaseEventEmitter()

    uplifted_ee = uplift(
        UpliftedEventEmitter,
        base_ee,
        proxy_new_listener=proxy_new_listener
    )

    @base_ee.on('new_listener')
    def base_new_listener_handler(event, f):
        assert event in ('event', 'new_listener')
        call_me('base new listener handler', f)

    @uplifted_ee.on('new_listener')
    def uplifted_new_listener_handler(event, f):
        assert event in ('event', 'new_listener')
        call_me('uplifted new listener handler', f)

    def fresh_base_handler():
        pass

    def fresh_uplifted_handler():
        pass

    base_ee.on('event', fresh_base_handler)
    uplifted_ee.on('event', fresh_uplifted_handler)

    if proxy_new_listener == 'both':
        call_me.assert_has_calls([
            call('base new listener handler', fresh_base_handler),
            call('uplifted new listener handler', fresh_base_handler),
            call('uplifted new listener handler', fresh_uplifted_handler),
            call('base new listener handler', fresh_uplifted_handler)
        ])
    elif proxy_new_listener == 'neither':
        call_me.assert_has_calls([
            call('base new listener handler', fresh_base_handler),
            call('uplifted new listener handler', fresh_uplifted_handler)
        ])
    elif proxy_new_listener == 'forward':
        call_me.assert_has_calls([
            call('base new listener handler', fresh_base_handler),
            call('uplifted new listener handler', fresh_base_handler),
            call('uplifted new listener handler', fresh_uplifted_handler)
        ])
    elif proxy_new_listener == 'backward':
        call_me.assert_has_calls([
            call('base new listener handler', fresh_base_handler),
            call('uplifted new listener handler', fresh_uplifted_handler),
            call('base new listener handler', fresh_uplifted_handler)
        ])
    else:
        raise Exception('unrecognized proxy_new_listener')



# -*- coding: utf-8 -*-

from asyncio import ensure_future, Future, iscoroutine
from pyee._base import BaseEventEmitter

__all__ = ['AsyncIOEventEmitter']


class AsyncIOEventEmitter(BaseEventEmitter):
    """An event emitter class which can run asyncio coroutines in addition to
    synchronous blocking functions. For example::

        @ee.on('event')
        async def async_handler(*args, **kwargs):
            await returns_a_future()

    On emit, the event emitter  will automatically schedule the coroutine using
    ``asyncio.ensure_future`` and the configured event loop (defaults to
    ``asyncio.get_event_loop()``).

    Unlike the case with the BaseEventEmitter, all exceptions raised by
    event handlers are automatically emitted on the ``error`` event. This is
    important for asyncio coroutines specifically but is also handled for
    synchronous functions for consistency.

    When ``loop`` is specified, the supplied event loop will be used when
    scheduling work with ``ensure_future``. Otherwise, the default asyncio
    event loop is used.

    For asyncio coroutine event handlers, calling emit is non-blocking.
    In other words, you do not have to await any results from emit, and the
    coroutine is scheduled in a fire-and-forget fashion.
    """
    def __init__(self, loop=None):
        super(AsyncIOEventEmitter, self).__init__()
        self._loop = loop

    def _emit_run(self, f, args, kwargs):
        try:
            coro = f(*args, **kwargs)
        except Exception as exc:
            self.emit('error', exc)
        else:
            if iscoroutine(coro):
                if self._loop:
                    f = ensure_future(coro, loop=self._loop)
                else:
                    f = ensure_future(coro)
            elif isinstance(coro, Future):
                f = coro
            else:
                f = None

            if f:
                @f.add_done_callback
                def _callback(f):
                    if f.cancelled():
                        return

                    exc = f.exception()
                    if exc:
                        self.emit('error', exc)



# -*- coding: utf-8 -*-

"""
pyee supplies a ``BaseEventEmitter`` class that is similar to the
``EventEmitter`` class from Node.js. In addition, it supplies the subclasses
``AsyncIOEventEmitter``, ``TwistedEventEmitter`` and ``ExecutorEventEmitter``
for supporting async and threaded execution with asyncio, twisted, and
concurrent.futures Executors respectively, as supported by the environment.


Example
-------

::

    In [1]: from pyee import BaseEventEmitter

    In [2]: ee = BaseEventEmitter()

    In [3]: @ee.on('event')
       ...: def event_handler():
       ...:     print('BANG BANG')
       ...:

    In [4]: ee.emit('event')
    BANG BANG

    In [5]:

"""

from pyee._base import (
    BaseEventEmitter,
    PyeeException
)
from pyee._compat import CompatEventEmitter as EventEmitter

__all__ = ['BaseEventEmitter', 'EventEmitter', 'PyeeException']

try:
    from pyee._asyncio import AsyncIOEventEmitter  # noqa
    __all__.append('AsyncIOEventEmitter')
except ImportError:
    pass

try:
    from pyee._twisted import TwistedEventEmitter  # noqa
    __all__.append('TwistedEventEmitter')
except ImportError:
    pass

try:
    from pyee._executor import ExecutorEventEmitter  # noqa
    __all__.append('ExecutorEventEmitter')
except ImportError:
    pass

try:
    from pyee._trio import TrioEventEmitter  # noqa
    __all__.append('TrioEventEmitter')
except (ImportError, SyntaxError):
    pass



# -*- coding: utf-8 -*-

from os import path
from setuptools import find_packages, setup

README_rst = path.join(path.abspath(path.dirname(__file__)), 'README.rst')

with open(README_rst, 'r') as f:
    long_description = f.read()

setup(
    name="pyee",
    vcversioner={},

    packages=find_packages(),
    setup_requires=[
        'pytest-runner',
        'pytest-asyncio; python_version > "3.4"',
        'pytest-trio; python_version > "3.7"',
        'vcversioner'
    ],
    tests_require=[
        'twisted',
        'trio; python_version > "3.6"',
        'futures; python_version < "3.0"'
    ],
    include_package_data=True,

    description="A port of node.js's EventEmitter to python.",
    long_description=long_description,
    author="Joshua Holbrook",
    author_email="josh.holbrook@gmail.com",
    url="https://github.com/jfhbrook/pyee",
    keywords=[
        "events", "emitter", "node.js", "node", "eventemitter",
        "event_emitter"
    ],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Other/Nonlisted Topic"
    ]
)



# -*- coding: utf-8 -*-

import pytest
import pytest_asyncio.plugin  # noqa

from asyncio import Future, wait_for

try:
    from asyncio.exceptions import TimeoutError
except ImportError:
    from concurrent.futures import TimeoutError

from mock import Mock

from twisted.internet.defer import ensureDeferred, succeed

from pyee import EventEmitter, AsyncIOEventEmitter, TwistedEventEmitter


class PyeeTestError(Exception):
    pass


@pytest.mark.parametrize('cls', [
    AsyncIOEventEmitter,
    EventEmitter
])
@pytest.mark.asyncio
async def test_asyncio_emit(cls, event_loop):
    """Test that asyncio-supporting event emitters can handle wrapping
    coroutines
    """

    ee = cls(loop=event_loop)

    should_call = Future(loop=event_loop)

    @ee.on('event')
    async def event_handler():
        should_call.set_result(True)

    ee.emit('event')

    result = await wait_for(should_call, 0.1)

    assert result is True


@pytest.mark.parametrize('cls', [
    AsyncIOEventEmitter,
    EventEmitter
])
@pytest.mark.asyncio
async def test_asyncio_once_emit(cls, event_loop):
    """Test that asyncio-supporting event emitters also wrap coroutines when
    using once
    """

    ee = cls(loop=event_loop)

    should_call = Future(loop=event_loop)

    @ee.once('event')
    async def event_handler():
        should_call.set_result(True)

    ee.emit('event')

    result = await wait_for(should_call, 0.1)

    assert result is True


@pytest.mark.parametrize('cls', [
    AsyncIOEventEmitter,
    EventEmitter
])
@pytest.mark.asyncio
async def test_asyncio_error(cls, event_loop):
    """Test that asyncio-supporting event emitters can handle errors when
    wrapping coroutines
    """
    ee = cls(loop=event_loop)

    should_call = Future(loop=event_loop)

    @ee.on('event')
    async def event_handler():
        raise PyeeTestError()

    @ee.on('error')
    def handle_error(exc):
        should_call.set_result(exc)

    ee.emit('event')

    result = await wait_for(should_call, 0.1)

    assert isinstance(result, PyeeTestError)


@pytest.mark.asyncio
async def test_asyncio_cancellation(event_loop):
    """Test that AsyncIOEventEmitter can handle Future cancellations"""

    cancel_me = Future(loop=event_loop)
    should_not_call = Future(loop=event_loop)

    ee = AsyncIOEventEmitter(loop=event_loop)

    @ee.on('event')
    async def event_handler():
        cancel_me.cancel()

    @ee.on('error')
    def handle_error(exc):
        should_not_call.set_result(None)

    ee.emit('event')

    try:
        await wait_for(should_not_call, 0.1)
    except TimeoutError:
        pass
    else:
        raise PyeeTestError()


@pytest.mark.asyncio
async def test_sync_error(event_loop):
    """Test that regular functions have the same error handling as coroutines
    """
    ee = AsyncIOEventEmitter(loop=event_loop)

    should_call = Future(loop=event_loop)

    @ee.on('event')
    def sync_handler():
        raise PyeeTestError()

    @ee.on('error')
    def handle_error(exc):
        should_call.set_result(exc)

    ee.emit('event')

    result = await wait_for(should_call, 0.1)

    assert isinstance(result, PyeeTestError)


@pytest.mark.parametrize('cls,kwargs', [
    (TwistedEventEmitter, dict()),
    (EventEmitter, dict(scheduler=ensureDeferred))
])
def test_twisted_emit(cls, kwargs):
    """Test that twisted-supporting event emitters can handle wrapping
    coroutines
    """
    ee = cls(**kwargs)

    should_call = Mock()

    @ee.on('event')
    async def event_handler():
        _ = await succeed('yes!')
        should_call(True)

    ee.emit('event')

    should_call.assert_called_once()


@pytest.mark.parametrize('cls,kwargs', [
    (TwistedEventEmitter, dict()),
    (EventEmitter, dict(scheduler=ensureDeferred))
])
def test_twisted_once(cls, kwargs):
    """Test that twisted-supporting event emitters also wrap coroutines for
    once
    """
    ee = cls(**kwargs)

    should_call = Mock()

    @ee.once('event')
    async def event_handler():
        _ = await succeed('yes!')
        should_call(True)

    ee.emit('event')

    should_call.assert_called_once()


def test_twisted_error():
    """Test that TwistedEventEmitters handle Failures when wrapping coroutines.
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on('event')
    async def event_handler():
        raise PyeeTestError()

    @ee.on('failure')
    def handle_error(e):
        should_call(e)

    ee.emit('event')

    should_call.assert_called_once()



# -*- coding: utf-8 -*-

from mock import Mock
from twisted.internet.defer import inlineCallbacks
from twisted.python.failure import Failure

from pyee import TwistedEventEmitter


class PyeeTestError(Exception):
    pass


def test_propagates_failure():
    """Test that TwistedEventEmitters can propagate failures
    from twisted Deferreds
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on('event')
    @inlineCallbacks
    def event_handler():
        yield Failure(PyeeTestError())

    @ee.on('failure')
    def handle_failure(f):
        assert isinstance(f, Failure)
        should_call(f)

    ee.emit('event')

    should_call.assert_called_once()


def test_propagates_sync_failure():
    """Test that TwistedEventEmitters can propagate failures
    from twisted Deferreds
    """
    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on('event')
    def event_handler():
        raise PyeeTestError()

    @ee.on('failure')
    def handle_failure(f):
        assert isinstance(f, Failure)
        should_call(f)

    ee.emit('event')

    should_call.assert_called_once()


def test_propagates_exception():
    """Test that TwistedEventEmitters propagate failures as exceptions to
    the error event when no failure handler
    """

    ee = TwistedEventEmitter()

    should_call = Mock()

    @ee.on('event')
    @inlineCallbacks
    def event_handler():
        yield Failure(PyeeTestError())

    @ee.on('error')
    def handle_error(exc):
        assert isinstance(exc, Exception)
        should_call(exc)

    ee.emit('event')

    should_call.assert_called_once()
